import uuid
from azure.identity import DefaultAzureCredential, get_bearer_token_provider
from openai import AsyncAzureOpenAI as async_aoi
from openai.types import CompletionUsage
from opentelemetry.trace import SpanKind

from foundationallm.langchain.agents import AgentBase
from foundationallm.langchain.exceptions import LangChainException
from foundationallm.langchain.language_models import LanguageModelFactory
from foundationallm.langchain.tools import ToolFactory
from foundationallm.langchain.workflows import WorkflowFactory
from foundationallm.models.agents import (
    GenericAgentWorkflow,
    AzureAIAgentServiceAgentWorkflow,
    AzureOpenAIAssistantsAgentWorkflow,
    LangChainAgentWorkflow,
    LangGraphReactAgentWorkflow,
    LangChainExpressionLanguageAgentWorkflow,
    ExternalAgentWorkflow
)
from foundationallm.models.constants import (
    ResourceObjectIdPropertyNames,
    ResourceObjectIdPropertyValues,
    ResourceProviderNames,
    AIModelResourceTypeNames,
    PromptResourceTypeNames
)
from foundationallm.models.operations import OperationTypes
from foundationallm.models.orchestration import (
    CompletionRequestObjectKeys,
    CompletionResponse
)
from foundationallm.models.resource_providers.ai_models import AIModelBase
from foundationallm.models.resource_providers.configuration import APIEndpointConfiguration
from foundationallm.models.agents import (
    AgentConversationHistorySettings,
    CompletionRequest
)
from foundationallm.models.attachments import AttachmentProviders
from foundationallm.models.authentication import AuthenticationTypes
from foundationallm.models.language_models import LanguageModelProvider
from foundationallm.models.resource_providers.prompts import MultipartPrompt
from foundationallm.models.services import OpenAIAssistantsAPIRequest
from foundationallm.services import (
    ImageService,
    OpenAIAssistantsApiService
)
from foundationallm.utils import ObjectUtils

from langchain_core.language_models import BaseLanguageModel

class GenericAgent(AgentBase):
    """
    The LangChain Knowledge Management agent.
    """

    def _validate_conversation_history(self, conversation_history_settings: AgentConversationHistorySettings):
        """
        Validates that the agent contains all required properties.

        Parameters
        ----------
        agent : KnowledgeManagementAgent
            The agent to validate.
        """
        if conversation_history_settings is None:
            raise LangChainException("The ConversationHistory property of the agent cannot be null.", 400)

        if conversation_history_settings.enabled is None:
            raise LangChainException("The Enabled property of the agent's ConversationHistory property cannot be null.", 400)

        if conversation_history_settings.enabled and conversation_history_settings.max_history is None:
            raise LangChainException("The MaxHistory property of the agent's ConversationHistory property cannot be null.", 400)

    def _validate_request(self, request: CompletionRequest):
        """
        Validates that the completion request contains all required properties.

        Parameters
        ----------
        request : KnowledgeManagementCompletionRequest
            The completion request to validate.
        """
        if request.agent is None:
            raise LangChainException("The agent property on the completion request cannot be null.", 400)

        if request.objects is None:
            raise LangChainException("The objects property on the completion request cannot be null.", 400)

        ai_model_object_id = request.agent.workflow.get_resource_object_id_properties(
            ResourceProviderNames.FOUNDATIONALLM_AIMODEL,
            AIModelResourceTypeNames.AI_MODELS,
            ResourceObjectIdPropertyNames.OBJECT_ROLE,
            ResourceObjectIdPropertyValues.MAIN_MODEL
        )
        if ai_model_object_id is None:
            raise LangChainException("The agent's workflow AI models requires a main_model.", 400)
        ai_model = ObjectUtils.get_object_by_id(ai_model_object_id.object_id, request.objects, AIModelBase)

        prompt_object_id = request.agent.workflow.get_resource_object_id_properties(
            ResourceProviderNames.FOUNDATIONALLM_PROMPT,
            PromptResourceTypeNames.PROMPTS,
            ResourceObjectIdPropertyNames.OBJECT_ROLE,
            ResourceObjectIdPropertyValues.MAIN_PROMPT
        )
        if prompt_object_id is None:
            raise LangChainException("The agent's workflow prompt object dictionary requires a main_prompt.", 400)
        prompt = ObjectUtils.get_object_by_id(prompt_object_id.object_id, request.objects, MultipartPrompt)

        if ai_model.endpoint_object_id is None or ai_model.endpoint_object_id == '':
            raise LangChainException("The AI model object provided in the request's objects dictionary is invalid because it is missing an endpoint_object_id value.", 400)
        if ai_model.deployment_name is None or ai_model.deployment_name == '':
            raise LangChainException("The AI model object provided in the request's objects dictionary is invalid because it is missing a deployment_name value.", 400)
        if ai_model.model_parameters is None:
            raise LangChainException("The AI model object provided in the request's objects dictionary is invalid because the model_parameters value is None.", 400)
        api_endpoint = ObjectUtils.get_object_by_id(ai_model.endpoint_object_id, request.objects, APIEndpointConfiguration)
        if api_endpoint.provider is None or api_endpoint.provider == '':
            raise LangChainException("The API endpoint object provided in the request's objects dictionary is invalid because it is missing a provider value.", 400)
        try:
            LanguageModelProvider(api_endpoint.provider)
        except ValueError:
            raise LangChainException(f"The LLM provider {api_endpoint.provider} is not supported.", 400)
        if api_endpoint.provider == LanguageModelProvider.MICROSOFT:
            # Verify the api_endpoint_configuration includes the api_version property for Azure OpenAI models.
            if api_endpoint.api_version is None or api_endpoint.api_version == '':
                raise LangChainException("The api_version property of the api_endpoint_configuration object cannot be null or empty.", 400)
        if api_endpoint.url is None or api_endpoint.url == '':
            raise LangChainException("The API endpoint object provided in the request's objects dictionary is invalid because it is missing a url value.", 400)
        if api_endpoint.authentication_type is None or api_endpoint.authentication_type == '':
            raise LangChainException("The API endpoint object provided in the request's objects dictionary is invalid because it is missing an authentication_type value.", 400)

        try:
            AuthenticationTypes(api_endpoint.authentication_type)
        except ValueError:
            raise LangChainException(f"The authentication_type {self.api_endpoint.authentication_type} is not supported.", 400)

        if prompt.prefix is None or prompt.prefix == '':
            raise LangChainException("The Prompt object provided in the request's objects dictionary is invalid because it is missing a prefix value.", 400)

        if isinstance(request.agent.workflow, AzureOpenAIAssistantsAgentWorkflow):
            if request.agent.workflow.assistant_id is None or request.agent.workflow.assistant_id == '':
                raise LangChainException("The AzureOpenAIAssistantsAgentWorkflow object provided in the request's agent property is invalid because it is missing an assistant_id value.", 400)

        self._validate_conversation_history(request.agent.conversation_history_settings)

    def _get_image_gen_language_model(self, api_endpoint_object_id, objects: dict) -> BaseLanguageModel:
        api_endpoint = ObjectUtils.get_object_by_id(api_endpoint_object_id, objects, APIEndpointConfiguration)
        scope = api_endpoint.authentication_parameters.get('scope', 'https://cognitiveservices.azure.com/.default')
        # Set up a Azure AD token provider.
        token_provider = get_bearer_token_provider(
            DefaultAzureCredential(exclude_environment_credential=True),
            scope
        )

        return async_aoi(
            azure_endpoint=api_endpoint.url,
            api_version=api_endpoint.api_version,
            azure_ad_token_provider=token_provider,
        )

    async def invoke_async(self, request: CompletionRequest) -> CompletionResponse:
        """
        Executes an async completion request.
        If a vector index exists, it will be queryied with the user prompt.

        Parameters
        ----------
        request : KnowledgeManagementCompletionRequest
            The completion request to execute.

        Returns
        -------
        CompletionResponse
            Returns a CompletionResponse with the generated summary, the user_prompt,
            generated full prompt with context and token utilization and execution cost details.
        """
        self._validate_request(request)

        agent = request.agent

        #----------------------------------------------------------------------
        # Legacy Azure OpenAI Assistants API agent workflow
        #----------------------------------------------------------------------

        if isinstance(agent.workflow, AzureOpenAIAssistantsAgentWorkflow):

            # Legacy image analysis

            ai_model_object_properties = request.agent.workflow.get_resource_object_id_properties(
                ResourceProviderNames.FOUNDATIONALLM_AIMODEL,
                AIModelResourceTypeNames.AI_MODELS,
                ResourceObjectIdPropertyNames.OBJECT_ROLE,
                ResourceObjectIdPropertyValues.MAIN_MODEL
            )
            ai_model_object_id = ai_model_object_properties.object_id
            prompt_object_properties = request.agent.workflow.get_resource_object_id_properties(
                ResourceProviderNames.FOUNDATIONALLM_PROMPT,
                PromptResourceTypeNames.PROMPTS,
                ResourceObjectIdPropertyNames.OBJECT_ROLE,
                ResourceObjectIdPropertyValues.MAIN_PROMPT
            )
            prompt_object_id = prompt_object_properties.object_id
            prompt = ObjectUtils.get_object_by_id(prompt_object_id, request.objects, MultipartPrompt)

            language_model_factory = LanguageModelFactory(request.objects, self.config)

            # Used by image analysis
            ai_model = ObjectUtils.get_object_by_id(ai_model_object_id, request.objects, AIModelBase)

            image_analysis_results = None
            image_analysis_token_usage = CompletionUsage(prompt_tokens=0, completion_tokens=0, total_tokens=0)
            # Get image attachments that are images with URL file paths.
            image_attachments = [attachment for attachment in request.attachments if (attachment.provider == AttachmentProviders.FOUNDATIONALLM_ATTACHMENT and attachment.content_type.startswith('image/'))] if request.attachments is not None else []
            if len(image_attachments) > 0:
                image_client = language_model_factory.get_language_model(ai_model_object_id, override_operation_type=OperationTypes.IMAGE_SERVICES)
                image_svc = ImageService(config=self.config, client=image_client, deployment_name=ai_model.deployment_name)
                image_analysis_results, usage = await image_svc.analyze_images_async(image_attachments)
                if usage is not None:
                    image_analysis_token_usage.prompt_tokens += usage.prompt_tokens
                    image_analysis_token_usage.completion_tokens += usage.completion_tokens
                    image_analysis_token_usage.total_tokens += usage.total_tokens

            # Legacy Assistants API implementation

            assistant_id = agent.workflow.assistant_id

            # create the service
            assistant_svc = OpenAIAssistantsApiService(
                azure_openai_client=language_model_factory.get_language_model(ai_model_object_id, override_operation_type=OperationTypes.ASSISTANTS_API),
                operations_manager=self.operations_manager
            )

            # populate service request object
            assistant_req = OpenAIAssistantsAPIRequest(
                document_id=str(uuid.uuid4()),
                operation_id=request.operation_id,
                instance_id=self.config.get_value("FoundationaLLM:Instance:Id"),
                assistant_id=assistant_id,
                thread_id=request.objects[CompletionRequestObjectKeys.OPENAI_THREAD_ID],
                attachments=[attachment.provider_file_name for attachment in request.attachments if attachment.provider == AttachmentProviders.FOUNDATIONALLM_AZURE_OPENAI],
                user_prompt=request.user_prompt
            )

            # Add user and assistant messages related to image analysis to the Assistants API request.
            if image_analysis_results is not None:
                # Add user message
                await assistant_svc.add_thread_message_async(
                    thread_id = assistant_req.thread_id,
                    role = "user",
                    content = "Analyze any attached images.",
                    attachments = []
                )
                # Add assistant message
                await assistant_svc.add_thread_message_async(
                    thread_id = assistant_req.thread_id,
                    role = "assistant",
                    content = image_svc.format_results(image_analysis_results),
                    attachments = []
                )

            image_service = None
            if any(tool.name == "DALLEImageGeneration" for tool in agent.tools):
                dalle_tool = next((tool for tool in agent.tools if tool.name == "DALLEImageGeneration"), None)

                model_object_id = dalle_tool.get_resource_object_id_properties(
                    ResourceProviderNames.FOUNDATIONALLM_AIMODEL,
                    AIModelResourceTypeNames.AI_MODELS,
                    ResourceObjectIdPropertyNames.OBJECT_ROLE,
                    ResourceObjectIdPropertyValues.MAIN_MODEL
                )

                image_generation_deployment_model = request.objects[model_object_id.object_id]["deployment_name"]
                api_endpoint_object_id = request.objects[model_object_id.object_id]["endpoint_object_id"]
                image_generation_client = self._get_image_gen_language_model(api_endpoint_object_id=api_endpoint_object_id, objects=request.objects)
                image_service=ImageService(
                    config=self.config,
                    client=image_generation_client,
                    deployment_name=image_generation_deployment_model,
                    image_generator_tool_description=dalle_tool.description)

            # invoke/run the service
            assistant_response = await assistant_svc.run_async(
                assistant_req,
                image_service=image_service
            )

            # Verify the Assistants API response
            if assistant_response is None:
                print("Assistants API response was None.")
                return CompletionResponse(
                    operation_id = request.operation_id,
                    full_prompt = prompt.prefix,
                    user_prompt = request.user_prompt,
                    user_prompt_rewrite = request.user_prompt_rewrite,
                    errors = [ "Assistants API response was None." ],
                    is_error = True
                )

            # create the CompletionResponse object
            return CompletionResponse(
                id = assistant_response.document_id,
                operation_id = request.operation_id,
                full_prompt = prompt.prefix,
                content = assistant_response.content,
                analysis_results = assistant_response.analysis_results,
                completion_tokens = assistant_response.completion_tokens + image_analysis_token_usage.completion_tokens,
                prompt_tokens = assistant_response.prompt_tokens + image_analysis_token_usage.prompt_tokens,
                total_tokens = assistant_response.total_tokens + image_analysis_token_usage.total_tokens,
                user_prompt = request.user_prompt,
                user_prompt_rewrite = request.user_prompt_rewrite,
                errors = assistant_response.errors,
                is_error = len(assistant_response.errors) > 0

            )
        # End Assistants API implementation

        #----------------------------------------------------------------------
        # Plugin-based agent workflows
        #----------------------------------------------------------------------

        if isinstance(agent.workflow, (
            GenericAgentWorkflow,
            AzureAIAgentServiceAgentWorkflow,
            LangChainAgentWorkflow,
            LangGraphReactAgentWorkflow,
            LangChainExpressionLanguageAgentWorkflow,
            ExternalAgentWorkflow)):

            # Ensure legacy agents do not require definition changes to use
            # the plugin-based workflows

            if isinstance(agent.workflow, (LangGraphReactAgentWorkflow)):
                agent.workflow.package_name = "foundationallm_agent_plugins_langchain"
                agent.workflow.class_name = "FoundationaLLMLangGraphReActAgentWorkflow"

            if isinstance(agent.workflow, (LangChainExpressionLanguageAgentWorkflow)):
                agent.workflow.package_name = "foundationallm_agent_plugins_langchain"
                agent.workflow.class_name = "FoundationaLLMLangChainLCELWorkflow"

            with self.tracer.start_as_current_span('langchain_prepare_plugin_workflow', kind=SpanKind.SERVER) as span:
                span.set_attribute("agent_name", agent.name)
                span.set_attribute("conversation_id", request.session_id if request.session_id is not None else '')
                span.set_attribute("operation_id", request.operation_id)

                # prepare tools
                tool_factory = ToolFactory(self.plugin_manager)
                workflow_tools = []

                # Populate tools list from agent configuration
                for tool_config in agent.tools:
                    tool_instance = \
                        tool_factory.get_tool(agent.name, tool_config, request.objects, self.user_identity, self.config)
                    tool_instance.description = tool_config.description
                    tool_instance.tool_config = tool_config
                    tool_instance.objects = request.objects
                    workflow_tools.append(tool_instance)

                request.objects['message_history'] = request.message_history[:agent.conversation_history_settings.max_history*2]

                # create the workflow
                workflow_factory = WorkflowFactory(
                    self.plugin_manager,
                    self.operations_manager)
                workflow = workflow_factory.get_workflow(
                    agent.workflow,
                    request.objects,
                    workflow_tools,
                    self.user_identity,
                    self.config)

            with self.tracer.start_as_current_span('langchain_invoke_plugin_workflow', kind=SpanKind.SERVER) as span:
                span.set_attribute("agent_name", agent.name)
                span.set_attribute("conversation_id", request.session_id if request.session_id is not None else '')
                span.set_attribute("operation_id", request.operation_id)

                response = await workflow.invoke_async(
                    operation_id=request.operation_id,
                    user_prompt=request.user_prompt,
                    user_prompt_rewrite=request.user_prompt_rewrite,
                    message_history=request.message_history,
                    file_history=request.file_history,
                    conversation_id=request.session_id,
                    objects=request.objects
                )
                # Ensure the user prompt rewrite is returned in the response
                response.user_prompt_rewrite = request.user_prompt_rewrite
            return response

        # The agent workflow type is not supported.

        return CompletionResponse(
            operation_id = request.operation_id,
            content = f'The agent workflow type {agent.workflow.type} is not supported.',
            user_prompt = request.user_prompt,
            user_prompt_rewrite = request.user_prompt_rewrite,
            full_prompt = self.full_prompt.text,
            completion_tokens = 0,
            prompt_tokens = 0,
            total_tokens = 0,
            total_cost = 0
        )



from types import ModuleType
from typing import Union, List

from .workflows.workflow_plugin_manager_base import WorkflowPluginManagerBase
from .tools.tool_plugin_manager_base import ToolPluginManagerBase

class ExternalModule():
    """
    Encapsulates properties useful for configuring an external module.
        module_file: str - The name of the module file.
        module_name: str - The name of the module.
        module_loaded: bool - Indicates whether the module is loaded.
        module: ModuleType - The module object.
        plugin_manager_class_names: List[str] - The list of plugin manager class names for the module.
        plugin_manager: List[Union[ToolPluginManager, WorkflowPluginManager]] - The list of plugin managers for the module.
    """

    module_file: str
    module_name: str
    module_loaded: bool = False
    module: ModuleType = None
    plugin_manager_class_names: List[str] = None
    plugin_managers: List[Union[ToolPluginManagerBase, WorkflowPluginManagerBase]] = None

    def __init__(self, module_file: str, module_name: str, plugin_manager_class_names: List[str]):
        """
        Initializes the external module.

        Parameters
        ----------
        module_file : str
            The name of the module file.
        module_name : str
            The name of the module.
        plugin_manager_class_name : List[Union[ToolPluginManager, WorkflowPluginManager]]
            The list of plugin managers for the module.
        """
        self.module_file = module_file
        self.module_name = module_name
        self.plugin_manager_class_names = plugin_manager_class_names
        self.plugin_managers = []



from .agent_workflow_ai_model import AgentWorkflowAIModel
from .agent_workflow_base import AgentWorkflowBase
from .azure_ai_agent_service_agent_workflow import AzureAIAgentServiceAgentWorkflow
from .azure_openai_assistants_agent_workflow import AzureOpenAIAssistantsAgentWorkflow
from .external_agent_workflow import ExternalAgentWorkflow
from .generic_agent_workflow import GenericAgentWorkflow
from .langchain_agent_workflow import LangChainAgentWorkflow
from .langchain_expression_language_agent_workflow import LangChainExpressionLanguageAgentWorkflow
from .langgraph_react_agent_workflow import LangGraphReactAgentWorkflow



from abc import abstractmethod
from foundationallm.config import Configuration, UserIdentity
from foundationallm.operations import OperationsManager
from foundationallm.models.orchestration import (
    CompletionRequestBase,
    CompletionResponse
)
from foundationallm.plugins import PluginManager
from foundationallm.telemetry import Telemetry

class AgentBase():
    """
    Implements the base functionality for a LangChain agent.
    """
    def __init__(self, instance_id: str, user_identity: UserIdentity, config: Configuration, plugin_manager: PluginManager, operations_manager: OperationsManager):
        """
        Initializes a knowledge management agent.

        Parameters
        ----------
        config : Configuration
            Application configuration class for retrieving configuration settings.
        """
        self.instance_id = instance_id
        self.user_identity = user_identity
        self.config = config
        self.plugin_manager = plugin_manager
        self.ai_model = None
        self.api_endpoint = None
        self.prompt = ''
        self.full_prompt = ''
        self.has_indexing_profiles = False
        self.has_retriever = False
        self.operations_manager = operations_manager

        self.tracer = Telemetry.get_tracer('langchain-agent-base')

    @abstractmethod
    async def invoke_async(self, request: CompletionRequestBase) -> CompletionResponse:
        """
        Gets the completion for the request using an async request.

        Parameters
        ----------
        request : CompletionRequestBase
            The completion request to execute.

        Returns
        -------
        CompletionResponse
            Returns a completion response.
        """
        raise NotImplementedError()

    @abstractmethod
    def _validate_request(self, request: CompletionRequestBase):
        """
        Validates that the completion request contains all required properties.

        Parameters
        ----------
        request : CompletionRequestBase
            The completion request to validate.
        """
        raise NotImplementedError()



from pydantic import BaseModel, Field

class UrlException(BaseModel):
    user_principal_name: str = Field(description="The user principal name.")
    url: str = Field(description="The alternative URL.")
    enabled: bool = Field(default=False, description="Whether the URL exception is enabled.")



from enum import Enum

class AttachmentProviders(str, Enum):
    """Enumerator of the Attachment Providers."""
    FOUNDATIONALLM_ATTACHMENT = "FoundationaLLM.Attachment"
    FOUNDATIONALLM_AZURE_OPENAI = "FoundationaLLM.AzureOpenAI"



from __future__ import annotations

from typing import TYPE_CHECKING

from .plugin_manager_types import PluginManagerTypes
from .plugin_manager import PluginManager

from .tools.tool_plugin_manager_base import ToolPluginManagerBase
from .workflows.workflow_plugin_manager_base import WorkflowPluginManagerBase



"""
Encapsulates properties useful for calling the OpenAI Assistants API.
"""
from pydantic import BaseModel
from typing import List, Optional

class OpenAIAssistantsAPIRequest(BaseModel):
    """
    Encapsulates properties useful for calling the OpenAI Assistants API.
        assistant_id: str - The ID of the assistant to use.
        thread_id: str - The ID of the conversation thread to use.
        attachments: Optional[List[AttachmentProperties]] - The list of OpenAI file attachments to include in the request.
        user_prompt: str - The user prompt/message to send to the assistants API.
    """
    document_id: Optional[str] = None
    operation_id: str
    instance_id: str
    assistant_id: str
    thread_id: str
    attachments: Optional[List[str]] = []
    user_prompt: str
    user_prompt_rewrite: Optional[str] = None



"""
Class: ToolFactory
Description: Factory class for creating tools based on the AgentTool configuration.
"""
import re

from foundationallm.config import Configuration, UserIdentity
from foundationallm.langchain.common import FoundationaLLMToolBase
from foundationallm.langchain.exceptions import LangChainException
from foundationallm.langchain.tools import DALLEImageGenerationTool
from foundationallm.models.agents import AgentTool
from foundationallm.plugins import PluginManager, PluginManagerTypes
from foundationallm.telemetry import Telemetry

class ToolFactory:
    """
    Factory class for creating tools based on the AgentTool configuration.
    """
    FLLM_PACKAGE_NAME = "FoundationaLLM"
    DALLE_IMAGE_GENERATION_TOOL = "DALLEImageGeneration"
    FOUNDATIONALLM_CONTENT_SEARCH_TOOL = "FoundationaLLMContentSearchTool"

    def __init__(self, plugin_manager: PluginManager):
        """
        Initializes the tool factory.

        Parameters
        ----------
        plugin_manager : PluginManager
            The plugin manager object used to load external tools.
        """
        self.plugin_manager = plugin_manager
        self.logger = Telemetry.get_logger(self.__class__.__name__)

    def get_tool(
        self,
        agent_name: str,
        tool_config: AgentTool,
        objects: dict,
        user_identity: UserIdentity,
        config: Configuration
    ) -> FoundationaLLMToolBase:
        """
        Creates an instance of a tool based on the tool configuration.
        """

        # NOTE: Disabled tool caching due to multiple implications with object lifetimes and stateful tools.
        # For example, tools that use LangChain LLM classed that authentication with tokens that expire pose
        # challenges when caching tools.

        # TODO: Revisit tool caching strategy in the future. Implement a more elaborate caching mechanism that
        # considers object lifetimes, stateful vs stateless tools, and token expiration handling.

        # Use a cache key based on agent name, package name, and tool name to store the tool instance in the object cache.
        # cache_key = f"{(self.__normalize_upn(user_identity.upn))}|{agent_name}|{tool_config.package_name}|{tool_config.name}"

        # if cache_key in self.plugin_manager.object_cache:
        #     self.logger.info("Using cached tool instance for key: %s", cache_key)
        #     return self.plugin_manager.object_cache[cache_key]

        # self.logger.info("Creating new tool instance for key: %s", cache_key)

        if tool_config.package_name == self.FLLM_PACKAGE_NAME:
            # Initialize by class name.
            match tool_config.class_name:
                case self.DALLE_IMAGE_GENERATION_TOOL:
                    tool = DALLEImageGenerationTool(tool_config, objects, user_identity, config)
                    # self.plugin_manager.object_cache[cache_key] = tool
            if tool is not None:
                return tool
        else:
            tool_plugin_manager = None

            if tool_config.package_name in self.plugin_manager.external_modules:
                tool_plugin_manager = next(( \
                    pm for pm \
                    in self.plugin_manager.external_modules[tool_config.package_name].plugin_managers \
                    if pm.plugin_manager_type == PluginManagerTypes.TOOLS), None)
                if tool_plugin_manager is None:
                    raise LangChainException(f"Tool plugin manager not found for package {tool_config.package_name}")
                tool = tool_plugin_manager.create_tool(tool_config, objects, user_identity, config)
                # self.plugin_manager.object_cache[cache_key] = tool
                return tool

            raise LangChainException(f"Package {tool_config.package_name} not found in the list of external modules loaded by the package manager.")

        raise LangChainException(f"Tool {tool_config.name} not found in package {tool_config.package_name}")

    def __normalize_upn(self, upn: str) -> str:
        """
        Normalize a UPN string by:
        - Trimming whitespace
        - Lowercasing
        - Replacing special characters with '-'
        """

        upn = upn.strip().lower()
        # Replace any character that is not a-z, 0-9 with '-'
        upn = re.sub(r'[^a-z0-9]', '-', upn)
        return upn


"""
Class: AzureAISearchServiceRetriever
Description: LangChain retriever for Azure AI Search.
"""
import json
from typing import List, Optional, Any
from langchain_openai import OpenAIEmbeddings
from langchain_core.callbacks import (
    AsyncCallbackManagerForRetrieverRun,
    CallbackManagerForRetrieverRun,
)
from langchain_core.documents import Document
from langchain_core.retrievers import BaseRetriever
from azure.search.documents import SearchClient
from azure.search.documents.models import VectorizedQuery
from azure.identity import DefaultAzureCredential
from foundationallm.models.orchestration import ContentArtifact
from foundationallm.models.vectors import VectorDocument
from foundationallm.services.gateway_text_embedding import GatewayTextEmbeddingService
from .content_artifact_retrieval_base import ContentArtifactRetrievalBase
from foundationallm.models.agents import KnowledgeManagementIndexConfiguration

class AzureAISearchServiceRetriever(BaseRetriever, ContentArtifactRetrievalBase):
    """
    LangChain retriever for Azure AI Search.
    Properties:
        config: Any -> Application configuration
        index_configurations: List[KnowledgeManagementIndexConfiguration]
            -> List of indexing profiles and associated API endpoint configurations
        gateway_text_embedding_service: GatewayTextEmbeddingService
            -> Service for retrieving text embeddings

    Searches embedding and text fields in the index for the top_n most relevant documents.

    Default FFLM document structure (overridable by setting the embedding and text field names):
        {
            "Id": "<GUID>",
            "Embedding": [0.1, 0.2, 0.3, ...], # embedding vector of the Text
            "Text": "text of the chunk",
            "Description": "General description about the source of the text",
            "AdditionalMetadata": "JSON string of metadata"
            "ExternalSourceName": "name and location the text came from, url, blob storage url"
            "IsReference": "true/false if the document is a reference document"
        }
    """
    config : Any
    index_configurations: List[KnowledgeManagementIndexConfiguration]
    gateway_text_embedding_service: GatewayTextEmbeddingService
    search_results: Optional[VectorDocument] = [] # Tuple of document id and document
    query_type: Optional[str] = "simple"
    semantic_configuration_name: Optional[str] = None
    top_n_override: Optional[int] = None
    
    def __get_embeddings(self, text: str) -> List[float]:
        """
        Returns embeddings vector for a given text.
        """
        embedding_response = self.gateway_text_embedding_service.get_embedding(text)
        return embedding_response.embedding_vector

    def _get_relevant_documents(
        self, query: str, *, run_manager: CallbackManagerForRetrieverRun
    ) -> List[Document]:
        """
        Performs a synchronous hybrid search on Azure AI Search index
        """

        self.search_results.clear()
        
        #search each indexing profile
        for index_config in self.index_configurations:

            credential_type = index_config.api_endpoint_configuration.authentication_type

            credential = None
            if credential_type == "AzureIdentity":
                credential = DefaultAzureCredential()

            endpoint = index_config.api_endpoint_configuration.url

            if self.top_n_override:
                top_n = self.top_n_override
            else:
                top_n = int(index_config.indexing_profile.settings.top_n)


            search_client = SearchClient(endpoint, index_config.indexing_profile.settings.index_name, credential)
            vector_query = VectorizedQuery(vector=self.__get_embeddings(query),
                                            k_nearest_neighbors=3,
                                            fields=index_config.indexing_profile.settings.embedding_field_name)

            results = search_client.search(
                search_text=query,
                filter=index_config.indexing_profile.settings.filters,
                vector_queries=[vector_query],
                query_type=self.query_type,
                semantic_configuration_name = self.semantic_configuration_name,
                top=top_n                
            )

            rerank_available = False

            #load search results into VectorDocument objects for score processing
            for result in results:
                metadata = {}              
                
                if index_config.indexing_profile.settings.metadata_field_name in result:
                    
                    try:
                        metadata = json.loads(result[index_config.indexing_profile.settings.metadata_field_name]) if index_config.indexing_profile.settings.metadata_field_name in result else {}
                    except Exception as e:
                        metadata = {}

                document = VectorDocument(
                        id=result[index_config.indexing_profile.settings.id_field_name],
                        page_content=result[index_config.indexing_profile.settings.text_field_name],
                        metadata=metadata,
                        score=result["@search.score"],
                        rerank_score=result.get("@search.reranker_score", 0.0)
                )
                if('@search.reranker_score' in result):                    
                    rerank_available = True
                    
                document.score = result["@search.score"]
                self.search_results.append(document)

        #sort search results by score
        if(rerank_available):
            self.search_results.sort(key=lambda x: (x.rerank_score, x.score), reverse=True)
        else:
            self.search_results.sort(key=lambda x: x.score, reverse=True)

        #take top n of search_results          
        self.search_results = self.search_results[:top_n]

        return self.search_results

    async def _aget_relevant_documents(
        self, query: str, *, run_manager: AsyncCallbackManagerForRetrieverRun
    ) -> List[Document]:
        """
        Performs an asynchronous hybrid search on Azure AI Search index
        NOTE: This functionality is not currently supported in the underlying Azure SDK.
        """
        raise Exception(f"Asynchronous search not supported.")

    def get_document_content_artifacts(self) -> List[ContentArtifact]:
        """
        Gets the content artifacts (sources) from the documents retrieved from the retriever.

        Returns:
            List of content artifacts (sources) from the retrieved documents.
        """
        content_artifacts = []
        added_ids = set()  # Avoid duplicates
        
        for result in self.search_results:  # Unpack the tuple
            result_id = result.id
            metadata = result.metadata
            if metadata is not None and 'multipart_id' in metadata and metadata['multipart_id']:
                if result_id not in added_ids:
                    title = (metadata['multipart_id'][-1]).split('/')[-1]
                    filepath = '/'.join(metadata['multipart_id'])
                    content_artifacts.append(ContentArtifact(id=result_id, title=title, filepath=filepath))
                    added_ids.add(result_id)
        return content_artifacts

    def format_docs(self, docs:List[Document]) -> str:
        """
        Generates a formatted string from a list of documents for use
        as the context for the completion request.
        """
        return "\n\n".join(doc.page_content for doc in docs)



"""
Class: FoundationaLLMWorkflowBase
Description: FoundationaLLM base class for tools that uses the agent workflow model for its configuration.
"""

from abc import ABC, abstractmethod
from typing import List, Optional

from azure.identity import DefaultAzureCredential
from langchain_core.messages import BaseMessage

from foundationallm.langchain.common import (
    FoundationaLLMToolBase
)
from foundationallm.config import (
    Configuration,
    UserIdentity
)
from foundationallm.langchain.language_models import LanguageModelFactory
from foundationallm.models.agents import (
    GenericAgentWorkflow,
    ExternalAgentWorkflow,
    AgentWorkflowBase
)
from foundationallm.models.constants import (
    AIModelResourceTypeNames,
    ContentArtifactTypeNames,
    PromptResourceTypeNames,
    ResourceObjectIdPropertyNames,
    ResourceObjectIdPropertyValues,
    ResourceProviderNames,
)
from foundationallm.models.messages import MessageHistoryItem
from foundationallm.models.orchestration import (
    CompletionResponse,
    ContentArtifact,
    FileHistoryItem
)
from foundationallm.models.resource_providers.ai_models import AIModelBase
from foundationallm.models.resource_providers.configuration import APIEndpointConfiguration
from foundationallm.operations import OperationsManager
from foundationallm.telemetry import Telemetry
from foundationallm.utils import LoggingAsyncHttpClient, ObjectUtils

class FoundationaLLMWorkflowBase(ABC):
    """
    FoundationaLLM base class for workflows that uses the agent workflow model for its configuration.
    """
    def __init__(
        self,
        workflow_config: GenericAgentWorkflow | ExternalAgentWorkflow | AgentWorkflowBase,
        objects: dict,
        tools: List[FoundationaLLMToolBase],
        operations_manager: OperationsManager,
        user_identity: UserIdentity,
        config: Configuration
    ):
        """
        Initializes the FoundationaLLMWorkflowBase class with the workflow configuration.

        Parameters
        ----------
        workflow_config : GenericAgentWorkflow | ExternalAgentWorkflow
            The workflow assigned to the agent.
        objects : dict
            The exploded objects assigned from the agent.
        tools : List[FoundationaLLMToolBase]
            The tools assigned to the agent.
        user_identity : UserIdentity
            The user identity of the user initiating the request.
        config : Configuration
            The application configuration for FoundationaLLM.
        """
        self.workflow_config = workflow_config
        self.objects = objects
        self.tools = tools if tools is not None else []
        self.operations_manager = operations_manager
        self.user_identity = user_identity
        self.config = config
        self.logger = Telemetry.get_logger(self.workflow_config.name)
        self.tracer = Telemetry.get_tracer(self.workflow_config.name)
        self.default_credential = DefaultAzureCredential(exclude_environment_credential=True)

        self.workflow_llm = None  # To be set in derived classes by calling create_workflow_llm()

    @abstractmethod
    async def invoke_async(
        self,
        operation_id: str,
        user_prompt: str,
        user_prompt_rewrite: Optional[str],
        message_history: List[MessageHistoryItem],
        file_history: List[FileHistoryItem],
        conversation_id: Optional[str] = None,
        objects: dict = None
    ) -> CompletionResponse:
        """
        Invokes the workflow asynchronously.

        Parameters
        ----------
        operation_id : str
            The unique identifier of the FoundationaLLM operation.
        user_prompt : str
            The user prompt message.
        user_prompt_rewrite : str
            The user prompt rewrite message containing additional context to clarify the user's intent.
        message_history : List[BaseMessage]
            The message history.
        file_history : List[FileHistoryItem]
            The file history.
        conversation_id : Optional[str]
            The conversation identifier for the workflow execution.
        objects : dict
            The exploded objects assigned from the agent. This is used to pass additional context to the workflow.
        """

    def create_workflow_execution_content_artifact(
        self,
        original_prompt: str,
        input_tokens: int = 0,
        output_tokens: int = 0,
        completion_time_seconds: float = 0,
        error_message: Optional[str] = None
    ) -> ContentArtifact:
        """
        Creates a content artifact for workflow execution.

        Parameters
        ----------
        original_prompt : str
            The original prompt used in the workflow execution.
        input_tokens : int
            The number of input tokens used. Defaults to 0.
        output_tokens : int
            The number of output tokens generated. Defaults to 0.
        completion_time_seconds : float
            The time taken for completion in seconds. Defaults to 0.
        error_message : Optional[str]
            An optional error message if an error occurred.

        Returns
        -------
        ContentArtifact
            The content artifact containing workflow execution details.
        """
        content_artifact = ContentArtifact(id=self.workflow_config.name)
        content_artifact.source = self.workflow_config.name
        content_artifact.type = ContentArtifactTypeNames.WORKFLOW_EXECUTION
        content_artifact.content = original_prompt
        content_artifact.title = self.workflow_config.name
        content_artifact.filepath = None
        content_artifact.metadata = {
            'prompt_tokens': str(input_tokens),
            'completion_tokens': str(output_tokens),
            'completion_time_seconds': str(completion_time_seconds)
        }
        if error_message:
            content_artifact.metadata['error_message'] = error_message
        return content_artifact

    def create_workflow_llm(
            self,
            intercept_http_calls: bool = False):
        """
        Creates the workflow LLM instance and saves it to self.workflow_llm.

        Parameters
        ----------
        intercept_http_calls : bool
            Whether to intercept HTTP calls for logging purposes. Defaults to False.

        Raises
        ------
        ValueError
            If no main model is found in the workflow configuration.
        """
        language_model_factory = LanguageModelFactory(self.objects, self.config)
        model_object_id = self.workflow_config.get_resource_object_id_properties(
            ResourceProviderNames.FOUNDATIONALLM_AIMODEL,
            AIModelResourceTypeNames.AI_MODELS,
            ResourceObjectIdPropertyNames.OBJECT_ROLE,
            ResourceObjectIdPropertyValues.MAIN_MODEL
        )
        if model_object_id:
            self.workflow_llm = \
                language_model_factory.get_language_model(
                    model_object_id.object_id,
                    http_async_client=LoggingAsyncHttpClient(timeout=30.0)
                ) if intercept_http_calls else \
                language_model_factory.get_language_model(
                    model_object_id.object_id
                )
        else:
            error_msg = 'No main model found in workflow configuration'
            self.logger.error(error_msg)
            raise ValueError(error_msg)

    def create_workflow_main_prompt(self) -> str:
        """
        Creates the workflow main prompt.

        Returns
        -------
        str
            The main prompt prefix from the workflow configuration.

        Raises
        ------
        ValueError
            If no main prompt is found in the workflow configuration.
        """
        prompt_object_id = self.workflow_config.get_resource_object_id_properties(
            ResourceProviderNames.FOUNDATIONALLM_PROMPT,
            PromptResourceTypeNames.PROMPTS,
            ResourceObjectIdPropertyNames.OBJECT_ROLE,
            ResourceObjectIdPropertyValues.MAIN_PROMPT
        )
        if prompt_object_id:
            main_prompt_object_id = prompt_object_id.object_id
            main_prompt_properties = self.objects[main_prompt_object_id]
            return main_prompt_properties['prefix']
        else:
            error_message = 'No main prompt found in workflow configuration'
            self.logger.error(error_message)
            raise ValueError(error_message)

    def create_workflow_router_prompt(self) -> str:
        """
        Creates the workflow router prompt.

        Returns
        -------
        str
            The router prompt prefix from the workflow configuration.

        Raises
        ------
        ValueError
            If no router prompt is found in the workflow configuration.
        """
        prompt_object_id = self.workflow_config.get_resource_object_id_properties(
            ResourceProviderNames.FOUNDATIONALLM_PROMPT,
            PromptResourceTypeNames.PROMPTS,
            ResourceObjectIdPropertyNames.OBJECT_ROLE,
            'router_prompt'
            # ResourceObjectIdPropertyValues.ROUTER_PROMPT
        )
        if prompt_object_id:
            router_prompt_object_id = prompt_object_id.object_id
            router_prompt_properties = self.objects[router_prompt_object_id]
            return router_prompt_properties['prefix']
        else:
            error_message = 'No router prompt found in workflow configuration'
            self.logger.error(error_message)
            raise ValueError(error_message)

    def create_workflow_files_prompt(self) -> str:
        """
        Creates the workflow files prompt.

        Returns
        -------
        str or None
            The files prompt prefix from the workflow configuration, or None if not found.
        """
        files_prompt_properties = self.workflow_config.get_resource_object_id_properties(
            ResourceProviderNames.FOUNDATIONALLM_PROMPT,
            PromptResourceTypeNames.PROMPTS,
            ResourceObjectIdPropertyNames.OBJECT_ROLE,
            ResourceObjectIdPropertyValues.FILES_PROMPT
        )
        if files_prompt_properties:
            files_prompt_object_id = files_prompt_properties.object_id
            return \
                self.objects[files_prompt_object_id]['prefix'] if files_prompt_object_id in self.objects else None
        else:
            warning_message = 'No files prompt found in workflow configuration'
            self.logger.warning(warning_message)
            return None

    def create_workflow_final_prompt(self) -> str:
        """
        Creates the workflow final prompt.

        Returns
        -------
        str or None
            The final prompt prefix from the workflow configuration, or None if not found.
        """
        final_prompt_properties = self.workflow_config.get_resource_object_id_properties(
            ResourceProviderNames.FOUNDATIONALLM_PROMPT,
            PromptResourceTypeNames.PROMPTS,
            ResourceObjectIdPropertyNames.OBJECT_ROLE,
            ResourceObjectIdPropertyValues.FINAL_PROMPT
        )
        if final_prompt_properties:
            final_prompt_object_id = final_prompt_properties.object_id
            return \
                self.objects[final_prompt_object_id]['prefix'] if final_prompt_object_id in self.objects else None
        else:
            warning_message = 'No final prompt found in workflow configuration'
            self.logger.warning(warning_message)
            return None

    def get_workflow_main_model_definition(
        self
    ) -> AIModelBase:
        """
        Gets the main model definition from the workflow configuration.

        Returns
        -------
        AIModelBase
            The AI model definition for the main model.
        """
        main_model_properties = self.workflow_config.get_resource_object_id_properties(
            ResourceProviderNames.FOUNDATIONALLM_AIMODEL,
            AIModelResourceTypeNames.AI_MODELS,
            ResourceObjectIdPropertyNames.OBJECT_ROLE,
            ResourceObjectIdPropertyValues.MAIN_MODEL
        )
        main_model_object_id = main_model_properties.object_id
        ai_model = ObjectUtils.get_object_by_id(main_model_object_id, self.objects, AIModelBase)
        return ai_model

    def get_ai_model_api_endpoint_configuration(
        self,
        ai_model: AIModelBase
    ) -> APIEndpointConfiguration:
        """
        Gets the API endpoint configuration for an AI model.

        Parameters
        ----------
        ai_model : AIModelBase
            The AI model to get the endpoint configuration for.

        Returns
        -------
        APIEndpointConfiguration
            The API endpoint configuration for the AI model.
        """
        api_endpoint = ObjectUtils.get_object_by_id(ai_model.endpoint_object_id, self.objects, APIEndpointConfiguration)
        return api_endpoint
    
    def get_text_from_message(self, message: BaseMessage) -> str:
        """
        Extracts text from content blocks returned by the LLM.

        Parameters
        ----------
        message : BaseMessage
            The message containing content blocks from the LLM response.

        Returns
        -------
        str
            The extracted text from all text-type content blocks, joined with spaces.
        """
        text_parts = [block["text"] for block in message.content_blocks if block.get("type") == "text"]
        text = " ".join(text_parts)
        return text.strip()



import aiohttp
import base64
from foundationallm.config import Configuration
from foundationallm.models.attachments import AttachmentProperties
from foundationallm.models.agents import CompletionRequest
from foundationallm.models.resource_providers.configuration import APIEndpointConfiguration
from foundationallm.storage import BlobStorageManager

class AudioAnalysisService:
    """
    Performs audio analysis services.
    """

    def __init__(self, config: Configuration):
        self.config = config
        
    def _get_as_base64(self, storage_account_name, file_path: str) -> str:
        """
        Retrieves an image from its URL and converts it to a base64 string.

        Parameters
        ----------
        mime_type : str
            The mime type of the image.
        image_url : str
            The URL of the image.

        Returns
        -------
        str
            The image as a base64 string.
        """
        try:
            # Remove any leading slashes from the file path.
            file_path = file_path.lstrip('/')
            # Attempt to retrieve the image from blob storage.
            container_name = file_path.split('/')[0]
            # Get the file path without the container name.
            file_name = file_path.removeprefix(container_name)

            try:
                storage_manager = BlobStorageManager(
                    account_name=storage_account_name,
                    container_name=container_name,
                    authentication_type=self.config.get_value('FoundationaLLM:ResourceProviders:Attachment:Storage:AuthenticationType')
                )
            except Exception as e:
                raise Exception(f'Error connecting to the {storage_account_name} blob storage account and the container named {container_name}: {e}')

            if (storage_manager.file_exists(file_name)):
                try:
                    # Get the image file from blob storage.
                   audio_blob = storage_manager.read_file_content(file_name)
                   return base64.b64encode(audio_blob).decode('utf-8')
                except Exception as e:
                    raise Exception(f'The specified image {storage_account_name}/{file_path} does not exist.')
            else:
                raise Exception(f'The specified image {storage_account_name}/{file_path} does not exist.')
        except Exception as e:
            print(f'Error getting image as base64: {e}')
            return None

    def _get_audio_classifcation_endpoint_configuration(self, request: CompletionRequest) -> APIEndpointConfiguration:
        """
        Gets the audio classification API endpoint configuration.

        Parameters
        ----------
        request : KnowledgeManagementCompletionRequest
            The request object.

        Returns
        -------
        APIEndpointConfiguration
            The audio classification API endpoint configuration.
        """
        api_configuration_object_id = request.agent.api_endpoint_configuration_object_ids['AudioClassificationAPI']
        api_configuration = request.objects.get(api_configuration_object_id, None)
        if api_configuration is None:
            raise Exception('Audio classification API endpoint configuration not found.')
        return APIEndpointConfiguration.from_object(api_configuration)

    async def classify_async(self, request: CompletionRequest, audio_attachments: AttachmentProperties):
        api_configuration = self._get_audio_classifcation_endpoint_configuration(request)
        api_key = self.config.get_value(api_configuration.authentication_parameters.get('api_key_configuration_name'))
        api_key_header_name = api_configuration.authentication_parameters.get('api_key_header_name', 'x-api-key')
        base_url = api_configuration.url.rstrip('/')
        deployment_name = api_configuration.properties.get('deployment_name')
        inference_endpoint = api_configuration.properties.get('inference_endpoint').lstrip('/')

        top_k = int(api_configuration.properties.get('top_k', 1))
        
        api_endpoint = f'{base_url}/{inference_endpoint}'
        audio_analyses = {}

        for attachment in audio_attachments:
            if attachment.content_type.startswith('audio/'):
                # Check for WAV and MP3 files only.
                if not attachment.content_type.endswith('wav') and not attachment.content_type.endswith('mp3'):
                    # TODO: Handle this properly.
                    continue

                audio_base64 = self._get_as_base64(storage_account_name=attachment.provider_storage_account_name, file_path=attachment.provider_file_name)
                if audio_base64 and audio_base64 != '':
                    # Create the payload, sending the base64 encoded audio file.
                    payload = {
                        "file": audio_base64,
                        "content_type": attachment.content_type,
                        "deployment_name": deployment_name,
                        "top_k": top_k
                    }
                    headers = {"charset": "utf-8", "Content-Type": "application/json", api_key_header_name: api_key}

                    # Make the REST API call.
                    async with aiohttp.ClientSession(headers=headers) as client:
                        async with client.post(api_endpoint, json=payload) as response:
                            if response.status != 200:
                                raise Exception(f'Error: ({response.status}) {response.text}')
                            
                            # Add the audio analysis to the dictionary.
                            audio_analyses[attachment.original_file_name] = await response.json()

        return audio_analyses
                    



from typing import Optional, List, Literal
from pydantic import Field
from .message_content_item_base import MessageContentItemBase
from .openai_file_path_message_content_item import OpenAIFilePathMessageContentItem
from .message_content_item_types import MessageContentItemTypes

class OpenAITextMessageContentItem(MessageContentItemBase):
    """An OpenAI text message content item."""

    type: Literal[MessageContentItemTypes.TEXT] = MessageContentItemTypes.TEXT
    agent_capability_category: Optional[str] = Field(None, alias="agent_capability_category")
    annotations: Optional[List[OpenAIFilePathMessageContentItem]] = Field(default_factory=list, alias="annotations")
    value: Optional[str] = Field(None, alias="value")



"""
Class Name: KnowledgeManagementCompletionRequest
Description: Encapsulates the metadata required to complete a knowledge management orchestration request.
"""
from typing import Optional

from foundationallm.models.orchestration import CompletionRequestBase

from .agent_base import AgentBase

class CompletionRequest(CompletionRequestBase):
    """
    The completion request received from the Orchestration API.
    """
    agent: Optional[AgentBase] = None
    objects: dict = {}
